(* --- Example of using the dependency analysis -------------------------- *)

(* This shows how we can use depdency analysis to mechanise the work of
 * moving theorems across from Nat to bNat. 
 * 
 * Our base will be the Peano axioms (nat_induct, Suc_inject, Suc_not_Zero)
 * and the definition of +. From this, 38 theorems (as of this writing) 
 * can be derived, and subsequently moved across.
 *)

(* Define the signature morphism from Nat to bNat  *)
val sg_n2b = SigMorph.make (theory "Nat") (theory "bNat")
               [("nat", SigMorph.Simple "bNat.bNat")]
	       [("0", "bNat.bNat.Zero"), ("Suc", "bNat.bSucc"), 
		("op +", "bNat.bPlus")];

(* Build the Nat dependency graph: *)
val dg= DepAna.get_dg (theory "Nat");

(* The base, and how it is translated in bNat. *)
val basemap=  
  [("Nat.nat_induct", "bNat.bNatInduct"),
   ("Nat.Suc_inject", "bNat.bSucc_inj"), 
   ("Nat.Suc_not_Zero", "bNat.succNotZero1"), 
   ("Nat.op +.add_0", "bNat.bPlus_Zero"),
   ("Nat.op +.add_Suc", "bNat.bPlus_bSucc")];

(* Let the dependency analysis figure out which of the theorems from
 * Nat can be derived from the base, and in which order. *)
val roadmap= DepAna.saturate dg (map fst basemap) 
                                (map fst (thms_of (theory "Nat")));

(* Now move them across: *)
AbsTac.abs_translate_base sg_n2b basemap roadmap;
