header {* The exception monad transformer *}

theory ExnT imports Main
begin 

use_thy Monad

typedecl 'a M


thymorph s : MonadType \<longrightarrow> ExnT 
maps [("'a MonadType.M" \<mapsto> "'a ExnT.M")] 


(* Get all renamed ops, axioms from monad and redefine the mixfix of bind_, 
   i.e. we employ an axiomatic instantiation here *)
t_instantiate MonadAxms mapping s 
renames: [("MonadOpEta.eta" \<mapsto> "eta'"), 
          ("MonadOpBind.bind" \<mapsto> "bind'" mixfix: ("_ >>=*\ _" [5, 6] 5) )] 
axiomatic


datatype 'a exn  = Error string | Just 'a
types 'a exnT = "('a exn) M" 


thymorph t : MonadType \<longrightarrow> ExnT
  maps [("'a M" \<mapsto> "'a exnT")] 
       

(* Here we need just both consts eta and bind.
   Due to instantiations above no renaming is required 
   and we get the standard mixfix for bind *)
t_instantiate MonadOps mapping t


(* Now we define the operations and prove the monad axioms *)
defs 
 eta_def:  "eta x == (eta' (Just x))"

defs
 bind_def: "(v >>= f) == (v >>=* (\<lambda>exn. case exn of Error e => (eta' (Error e))
                                                          | Just  x => (f x)))"

(* for exnT we instantiate without axioms ... *)
t_instantiate MonadAxms mapping t'


(* proving goals ... *)

lemma exnT_lunit:  "(eta x >>= t) = t x" 
by(simp add: bind_def eta_def mon_lunit)

lemma exnT_runit:  "(t >>= eta) = t"
proof -
 have "exn_case (\<lambda> e. eta' (Error e)) eta = eta'" (is "?L = ?R")
 proof (rule ext)
  fix x :: "'b exn"  
  show "?L x = ?R x"
  proof (cases x)
   case Error thus ?thesis by simp
  next
   case Just thus ?thesis by (simp add: eta_def)
  qed
 qed
 hence "(t >>=* ?L) = (t >>=* ?R)" by (rule arg_cong)
 also have "... = t" by (rule mon_runit)
 finally show ?thesis by (simp add: bind_def)
qed


lemma exnT_assoc:  "(s >>= t >>= u) = (s >>= (\<lambda>x. t x >>= u))" (is "?L = ?R")
proof -
 let ?L' = "(s >>=* exn_case (\<lambda>e. eta' (Error e)) t >>=* exn_case (\<lambda>e. eta' (Error e)) u)"
 have "?L = ?L'" by(simp add: bind_def eta_def)
 also have "... = (s >>=* (\<lambda>x. (exn_case (\<lambda>e. eta' (Error e)) t) x >>=* 
                                        exn_case (\<lambda>e. eta' (Error e)) u))"
 by (rule mon_assoc)
 also have "... = ?R"
  apply (simp add: bind_def eta_def, rule_tac f="\<lambda>x. s >>=* x" in arg_cong, rule ext)
  apply(case_tac x)
  apply(simp add: mon_lunit)
  by simp
 finally show ?thesis by assumption
qed


lemma exnT_inj: "eta x= eta y ==> x= y" 
proof(simp add: eta_def)
 assume "eta' (Just x) = eta' (Just y)"
 hence "Just x = Just y" by (rule mon_eta_inj)
 thus "x = y" by simp
qed

thymorph t by_sigmorph t''


(* Instantiate the exnT monad *)
t_instantiate Monad mapping t 

translate_thm "Monad.mon_assoc2" as "" along t'
translate_thm "Monad.lift_eta" as "" along t'
translate_thm "Monad.lift2_eta" as "" along t'
translate_thm "Monad.lift_eta2" as "" along t'
translate_thm "Monad.mapF_joinM" as "" along t'


text{* Raising and handling errors *}

constdefs
  raise  :: "string => 'a exnT"
 "raise err == (eta' (Error err))" 

constdefs
  handle :: "'a exnT => (string=> 'a M)=> 'a M"
 "handle mex f == (mex >>=* (\<lambda> ex. case ex of Error e => f e 
                                            | Just v  => eta' v))"
 
end
