
theory Squareroot 
imports Trafo
begin

subsection {* Auxiliary theorems about order *}

lemma fin_n : "finite ({x. x \<le> (n::nat)})" (is "finite (?S n)")
proof (induct n)
 show "finite (?S 0)" by simp
next
 fix n
 assume "finite (?S n)"
 thus "finite (?S(Suc n))"
 proof -
  have "?S (Suc n) = insert (Suc n) (?S n)" 
  proof (rule equalityI)
   show "?S (Suc n) \<le> insert (Suc n) (?S n)"
   by (rule subsetI, simp, rule disjCI, simp)
  next
   show "insert (Suc n) (?S n) \<le> ?S (Suc n) "
   by (rule subsetI, simp, erule disjE, simp_all)
  qed
  with prems show "finite (?S(Suc n))" by simp
 qed
qed

lemma fin_prod : 
 "\<lbrakk> finite S; finite T \<rbrakk> \<Longrightarrow> finite (S <*> T)" by simp

lemma fin_np : "finite ({(a, b). a \<le> (n::nat) \<and> b \<le> n})"
proof (simp, rule fin_prod)
 show "finite {a. a <= n}" by (rule fin_n)
next
 show "finite {b. b <= n}" by (rule fin_n) 
qed


section {* Stepwise construction of t : SplitImpl --> Squareroot *}

thymorph i : Main --> Squareroot

t_instantiate SplitSpecs mapping i
inst_types: [("D" \<mapsto> "nat"),
             ("R" \<mapsto> "nat")]


constdefs
 sqr :: "nat \<Rightarrow> nat"
"sqr x \<equiv> x*x"

defs
 Inv_def  : "Inv \<equiv> (\<lambda> (n, k). sqr k \<le> n)"
 Term_def : "Term \<equiv> (\<lambda> (n, k). n < sqr (Suc k))"

t_instantiate SplitOps mapping i'
(* Get the auxiliary operations. These we define as follows: *) 
defs
E_def     : "E n \<equiv> 0"

H_def     : "H \<equiv> (\<lambda> (n, k). Suc k)"

R_ord_def : "R_ord n \<equiv> {(a, b). (a \<le> n) \<and> (b \<le> n) \<and> (b < a)} "


(* A proper instantiation of SplitAxms *) 
t_instantiate SplitAxms mapping i''

(* Or alternatively: 
sigmorph s : SplitAxms --> Squareroot
maps [("D" \<mapsto> "nat"),
      ("R" \<mapsto> "nat")]
     [("SplitOps.E" |-> "Squareroot.E"),
      ("SplitSpecs.Inv" |-> "Squareroot.Inv"),
      ("SplitSpecs.Term" |-> "Squareroot.Term")]
show_thymorph_goals s
*)

subsection {* Proving the properties *}

lemma S1 : "\<forall> x. Inv (x, E x)"
by(simp add: Inv_def E_def sqr_def)

lemma S2 : "\<forall> x z. Inv (x, z) \<longrightarrow> (\<not> Term (x, z)) \<longrightarrow> Inv (x, H (x, z))"
by(simp add: Inv_def Term_def H_def)


lemma finR : " finite (R_ord n) "
by (simp add: R_ord_def, rule_tac B="{(a, b). a \<le> n \<and> b \<le> n}" in finite_subset,
    rule subsetI, simp add: split_def, rule fin_np)


lemma acycR : " acyclic (R_ord n)"
proof (simp add: acyclic_def, rule allI)
 fix x show "(x, x) \<notin> (R_ord n)^+"
 proof (rule ccontr, simp)
   assume "(x, x) \<in> (R_ord n)^+"
     have c: "\<forall> a b. (a, b) \<in> (R_ord n)^+ \<longrightarrow> b < a"
       by (intro allI, rule impI, simp add: R_ord_def, erule trancl_induct, simp_all)
     have "x < x" by (rule c[rule_format])
   thus "False" ..
 qed
qed


lemma Ord1 : "\<forall> b. wf (R_ord b)"
by (rule allI, rule finite_acyclic_wf, rule finR, rule acycR)


lemma Ord2 : "\<forall> y x. (\<not> Term (x, y)) \<longrightarrow> (H (x, y), y) \<in> (R_ord x) " 
by(simp add: Term_def H_def sqr_def R_ord_def)



(* Now the theory morphism from SplitAxms can be defined *)
thymorph t by_sigmorph i'''


text {* Get finally the recursive definition of g ... *} 
t_instantiate SplitImpl mapping t renames: [("SplitImpl.f" \<mapsto> "sqrt")]


text {* ... and the property of sqrt *}
translate_thm "SplitImpl.PostCond" as sqrt_prop along t'

end
