
(* Alpha-conversion for unvarified types and terms. 
*) 


signature ALPHA =
sig

type typevar_map

val rep_typevar_map : typevar_map -> string
val typ_aconv_map : Type.tsig -> (Term.typ * Term.typ) -> typevar_map option
val typ_aconv : Type.tsig -> (Term.typ * Term.typ) -> bool
val term_aconv : Type.tsig -> (Term.term * Term.term) -> bool
val map_aconv_typ : typevar_map -> Term.typ -> Term.typ
val map_aconv_term : typevar_map -> Term.term -> Term.term

end


structure Alpha : ALPHA =
struct


(* *** alpha-conversion for types and terms *** *)


type typevar_map = (string Symtab.table * (string * Term.sort) Symtab.table) * 
                   (string Symtab.table * string Symtab.table)

fun show_alpha_map mp = commas (map (fn (a, b) => (a ^" |-> "^ b)) mp) 

fun rep_typevar_map ((ty, ity), (mp, imp)) = (show_alpha_map (Symtab.dest ty) ^ show_alpha_map (Symtab.dest mp))



fun typ_aconv_map_ tsig ((ty, ity), (mp, imp)) (TFree (v, s)) (Type (f, [])) = 
   (case (Symtab.lookup ty v) of
     SOME f' => if (f' = f) then SOME ((ty, ity), (mp, imp))
                else NONE
   | NONE => (case (Symtab.lookup ity f) of
               SOME (v', s') => (if (v' = v andalso (Type.eq_sort tsig (s', s)))
                                 then SOME ((ty, ity), (mp, imp))
                                 else NONE)
             | NONE => (case (Symtab.lookup mp v) of
                          SOME _ => NONE
                        | NONE => (case (Symtab.lookup imp v) of
                                     SOME _ => NONE
                                   | NONE => 
                          SOME ((Symtab.update_new (v, f) ty, Symtab.update_new (f, (v, s)) ity), (mp, imp))))))

  | typ_aconv_map_ tsig ((ty, ity), (mp, imp)) (Type (f, [])) (TFree (v, s)) =
      typ_aconv_map_ tsig ((ty, ity), (mp, imp)) (TFree (v, s)) (Type (f, []))

  | typ_aconv_map_ tsig ((ty, ity), (mp, imp)) (TFree (v, s)) (TFree (w, t)) =
   (if (Type.eq_sort tsig (s, t)) then
   (case (Symtab.lookup ty v) of
     SOME _ => NONE
   | NONE => SOME ((ty, ity), (Symtab.insert (op =) (v, w) mp, Symtab.insert (op =) (w, v) imp))
             handle DUP => NONE)
    else NONE)

  | typ_aconv_map_ tsig ((ty, ity), (mp, imp)) (Type (f, ts)) (Type (f', ts')) =
    (if ((f = f') andalso (length ts = length ts')) 
     then typ_aconv_map__ tsig ((ty, ity), (mp, imp)) ts ts'
     else NONE)

  | typ_aconv_map_ tsig ((ty, ity), (mp, imp)) _ _ = NONE

and  
    typ_aconv_map__ tsig maps [] _ = SOME maps
  | typ_aconv_map__ tsig maps (t::ts) (t'::ts') = 
     (case (typ_aconv_map_ tsig maps t t') of
       NONE => NONE
     | SOME maps' => typ_aconv_map__ tsig maps' ts ts')


fun typ_aconv_map tsig (T, U) = 
    typ_aconv_map_ tsig ((Symtab.empty, Symtab.empty), (Symtab.empty, Symtab.empty)) T U
(* let val (T', U') = (Aux.unvarifyT T, Aux.unvarifyT U) *)





fun typ_aconv_ tsig (mp, imp) (TFree (v, s)) (TFree (w, t)) =
   (if (Type.eq_sort tsig (s, t)) then
     SOME (Symtab.insert (op =) (v, w) mp, Symtab.insert (op =) (w, v) imp)
     handle DUP => NONE
    else NONE)

  | typ_aconv_ tsig (mp, imp) (Type (f, ts)) (Type (f', ts')) =
     (if ((f = f') andalso (length ts = length ts')) 
      then typ_aconv__ tsig (mp, imp) ts ts'
      else NONE)

  | typ_aconv_ tsig (mp, imp) _ _ = NONE

and
    typ_aconv__ tsig maps [] [] = SOME maps
  | typ_aconv__ tsig maps (t::ts) (t'::ts') = 
     (case (typ_aconv_ tsig maps t t') of
       NONE => NONE
     | SOME maps' => typ_aconv__ tsig maps' ts ts')
  | typ_aconv__ tsig maps _ _ = NONE

fun typ_aconv tsig (T, U) = (case (typ_aconv_ tsig (Symtab.empty, Symtab.empty) T U) of
                               SOME _ => true
                             | NONE => false)



fun term_aconv_ tsig (mp, imp) (Const (c, tp)) (Const (c', tp')) = 
    (if ((c = c') andalso (typ_aconv tsig (tp, tp'))) 
     then SOME (mp, imp) else NONE)
  | term_aconv_ tsig (mp, imp) (Free (v, tp)) (Free (v', tp')) = 
    (if (typ_aconv tsig (tp, tp')) 
     then SOME (Symtab.insert (op =) (v, v') mp, Symtab.insert (op =) (v', v) imp)
          handle DUP => NONE
     else NONE) 
  | term_aconv_ tsig (mp, imp) (Var (ind, tp)) (Var (ind', tp')) = error ("Unexpected Var in term_aconv")
  | term_aconv_ tsig (mp, imp) (Bound i) (Bound j) = (if (i = j) then SOME (mp, imp) else NONE)
  | term_aconv_ tsig (mp, imp) (Abs(_, tp, tt)) (Abs(_, tp', tt')) = 
    (if (typ_aconv tsig (tp, tp')) then term_aconv_ tsig (mp, imp) tt tt'
                                   else NONE) 
  | term_aconv_ tsig (mp, imp) (tt $ ts) (tt' $ ts') = 
    (case (term_aconv_ tsig (mp, imp) tt tt') of
      SOME (mp', imp') => term_aconv_ tsig (mp', imp') ts ts'
    | NONE => NONE)
  | term_aconv_ tsig (mp, imp) _ _ = NONE


(* Decides whether argument terms t and s are alpha-convertible, i.e. there exists a bijection
   between Free vars  *)
fun term_aconv tsig (t, s) = (case (term_aconv_ tsig (Symtab.empty, Symtab.empty) t s) of
                                SOME _ => true
                              | NONE => false)





(* *** Convert a given type by a given typevar_map for free type vars (TFree). 
       The argument type has to be unvarified. 
   *** *) 
fun map_aconv_typ ((t_map, c_map), (mp, imp)) (TFree (v, s)) =
                                       (case (Symtab.lookup t_map v) of
                                          SOME f => (Type (f, []))
                                        | NONE => 
                                          (case (Symtab.lookup mp v) of
                                                NONE => (case (Symtab.lookup imp v) of
                                                          SOME v' => TFree (v', s)
                                                        | NONE => TFree (v, s))
                                             | SOME v' => TFree (v', s)))
  | map_aconv_typ ((t_map, c_map), (mp, imp)) (Type (f, [])) =
                                       (case (Symtab.lookup c_map f) of
                                         SOME (v, s) => TFree (v, s)
                                       | NONE => (Type (f, [])))  
  | map_aconv_typ (mp, imp) (TVar (i, s)) = error ("Unexpected schematic typevar in map_aconv_typ")
  | map_aconv_typ (mp, imp) (Type (f, ts)) = Type (f, map (map_aconv_typ (mp, imp)) ts)      

   
(* *** Convert all types in a given term by a given alpha-map for free type vars (as above) *** *)          
fun map_aconv_term mp (Const (c, tp)) = Const(c, map_aconv_typ mp tp)
  | map_aconv_term mp (Free (v, tp)) = Free(v, map_aconv_typ mp tp)
  | map_aconv_term mp (Var(i, tp)) = Var(i, map_aconv_typ mp tp)
  | map_aconv_term mp (Bound i) = (Bound i)
  | map_aconv_term mp (Abs (x, tp, tt)) = Abs(x, map_aconv_typ mp tp, map_aconv_term mp tt)
  | map_aconv_term mp (s $ t) = (map_aconv_term mp s) $ (map_aconv_term mp t)



end


