(* **************************************************************************
 *
 * Various auxiliary functions.
 *
 * $Source: /home/uniform/awe/repository/src/Extensions/auxiliary.sml,v $
 * $Revision: 940 $ of $Date: 2007-12-14 09:54:50 +0100 (Fr, 14 Dez 2007) $
 *)

structure Aux =

struct

(* ************************************************************* *)
(*                    Auxiliaries                                *)
(* ************************************************************* *)

fun warning s = Output.warning s
fun error_msg s = Output.error_msg s
fun info s = Output.info s 


(* From Isabelle2004 Library *)

(*association list lookup, optimized version for strings*)
fun assoc_string ([], (key:string)) = NONE
  | assoc_string ((keyi, xi) :: pairs, key) =
      if key = keyi then SOME xi else assoc_string (pairs, key);
fun assoc_int ([], (key:int)) = NONE
  | assoc_int ((keyi, xi) :: pairs, key) =
      if key = keyi then SOME xi else assoc_int (pairs, key);

fun foldl f (e, []) = e
  | foldl f (e, x::xs) = foldl f (f (e, x), xs)
fun foldr f ([], e) = e
  | foldr f (x::xs, e) = f(x, foldr f (xs, e))


fun flat (ls: 'c list list) : 'c list = foldl (op @) ([], ls);

fun overwrite (al, p as (key, _)) =
  let fun over ((q as (keyi, _)) :: pairs) =
            if keyi = key then p :: pairs else q :: (over pairs)
        | over [] = [p]
  in over al end;

val mapfilter = Library.mapfilter

val thms_of = PureThy.thms_of
val all_thms_of = PureThy.all_thms_of
val sign_of = Theory.sign_of

val warning_fn = ref (std_output o suffix "\n" o prefix_lines "### ");
val error_fn = ref (std_output o suffix "\n" o prefix_lines "*** ");
fun warning s = ! warning_fn s;
fun error_msg s = ! error_fn s;
fun error s = (error_msg s; raise ERROR);

fun seq (proc: 'a -> unit) : 'a list -> unit =
  let fun seqf [] = ()
        | seqf (x :: xs) = (proc x; seqf xs)
  in seqf end;




(* ***** Operations and logical types of a signature ***** *)

fun get_ops sg = Symtab.keys (snd(fst(#consts (Sign.rep_sg sg))))
fun get_ltypes sg = Type.logical_types (Sign.tsig_of sg)

fun type_arity(sg, tycon) = 
 let val rep = Type.rep_tsig (Sign.tsig_of sg)
 in
  case (Symtab.lookup (#arities rep) tycon) of
   SOME a => (case a of 
               [] => NONE
             | _ =>  SOME (a, length (snd(hd a))))
 | NONE => NONE
 end

(*
fun unvarifyT ty = Type.unvarifyT ty
fun unvarify t = Logic.unvarify t
*)

fun unvarifyT (TVar ((v, i), s)) = TFree (v, s)
  | unvarifyT (Type (f, ts)) = Type (f, map unvarifyT ts)
  | unvarifyT t = t 

fun unvarify (Const (a, T)) = Const (a, unvarifyT T)
  | unvarify (Free (a, T)) = Free (a, unvarifyT T)
  | unvarify (Var ((a, x), T)) = Free (a, unvarifyT T)
 (* | unvarify (Var (ixn, T)) = Var (ixn, unvarifyT T) *) (*non-0 index!*)
  | unvarify (t as Bound _) = t
  | unvarify (Abs (a, T, body)) = Abs (a, unvarifyT T, unvarify body)
  | unvarify (f $ t) = unvarify f $ unvarify t;


val const_type = Sign.const_type

(* ******************************************************* *)



fun is_dot #"." = true
  | is_dot   _  = false

fun has_dot s = (let val l = explode s
                     fun chk [] = false
                       | chk (x::xs) = (x = ".") orelse (chk xs)
                 in
                  chk l
                 end)



fun show_opmap (s, (snm, tnm)) = (s ^ "\nFound: (\"" ^ snm ^ "\" |-> \"" ^ tnm ^ "\")" )
fun show_amb (s, (nm, nmlist)) = (s ^ "\n\"" ^ nm ^ "\" |-> [ " ^ (commas nmlist) ^ " ]")


(* Name of theory : *)
(* fun name_of_thy thy = Sign.name_of (sign_of thy) *)
(* Must be changed because of stale signature exception for Isar transitions *)
fun name_of_thy thy = NameSpace.path_of(Sign.naming_of thy)
fun dot_name thy = (name_of_thy thy)^ "."

fun drop_base nm = (let val b = NameSpace.base nm
                    in
                     case (try (unsuffix ("." ^ b)) nm) of
                       NONE => nm
                     | SOME x => x 
                    end)
val pack = space_implode "."
val unpack = space_explode "."

fun gen_arity 0 = []
  | gen_arity n = (let val c = ("'" ^ (chr (96 + n)))
                   in
                    c :: (gen_arity (n - 1))
                   end)

(* **** *)

fun qname_split nm =
             let val s = explode nm
                 fun f qn [] = (qn, [])
                   | f qn (c::cs) = (if (c=".") then (qn, cs) else (f (qn@[c]) cs))
                 fun repl_pnts [] = []
                   | repl_pnts (c::cs) = (if (c=".") then ("_"::(repl_pnts cs))
                                          else (c::(repl_pnts cs)))
                 val (qn, n) = f [] s
             in
              case n of
               [] => ("", implode qn)
              | _ =>  (implode qn, implode (repl_pnts n))
             end

fun qname_split2 nm =
             let val s = explode nm
                 fun f qn [] = (qn, [])
                   | f qn (c::cs) = (if (c=".") then (qn, cs) else (f (qn@[c]) cs))
                 val (qn, n) = f [] s
             in
              case n of
               [] => ("", implode qn)
              | _ =>  (implode qn, implode n)
             end

fun string_subst str pat pat' =
 let val m = length (explode pat)
     fun hash_f p = foldr (fn (a, b) => (ord a + b)) (p, 0)
     val hash = hash_f (explode pat)
     fun take_n n [] = (if (n = 0) then (SOME ([], [])) else NONE)
       | take_n n (x::xs) =  (if (n = 0) then (SOME ([], x::xs))
                              else (case (take_n (n-1) xs) of
                                      NONE => NONE
                                    | (SOME (a, b)) => (SOME (x::a, b))))

     fun step_1 h w [] = (if (h = hash) then
                          (if ((implode w) = pat) then (explode pat')
                           else w)
                          else w)
       | step_1 h w (x::xs) = (if (h = hash) then
                               (if ((implode w) = pat) then ((explode pat')@(step_2 (x::xs)))
                                else ((hd w)::(step_1 (h - (ord (hd w)) + (ord x)) ((tl w) @ [x]) xs)))
                               else ((hd w)::(step_1 (h - (ord (hd w)) + (ord x)) ((tl w) @ [x]) xs)))
 
     and step_2 xs = (case (take_n m xs) of
                       NONE => xs
                     | SOME (w, xs') => step_1 (hash_f w) w xs') 
     
 in
  implode(step_2 (explode str)) 
 end                                     
  





infix 7 >>=
fun (opt >>= f) = case opt of
                    NONE => NONE
                  | SOME x => f x



(* For unvarified def. terms *)
fun abs_def (Const ("==>", t) $ A $ B) = (Const("==>", t) $ A $ (abs_def B))
  | abs_def term =
 let val opt = (try Logic.dest_equals) (Logic.strip_imp_concl term) >>= 
               (fn (lhs, rhs) => SOME (Term.strip_comb lhs, rhs))
 in
  case opt of
    NONE => error ("Only definitions can be decomposed")
  | SOME ((head, args), rhs) => 
                        (let val args' = rev args
                             fun abs t [] = t
                               | abs t (arg::xs) = 
                                 (case arg of
                                   Free(v, ty) => (let val t' = Term.abstract_over (arg, t)
                                                       val t'' = Abs(v, ty, t')
                                                   in
                                                    abs t'' xs
                                                   end)
                                 | _ => error "Unexpected argument form")
                          in
                            Logic.mk_equals (head, (abs rhs args'))
                          end)
 end



fun get_opname tm = 
    (try Logic.dest_equals) (Logic.strip_imp_concl tm) >>= 
    (fn (lhs, rhs) => SOME (Term.strip_comb lhs)) >>= 
    (fn (head, args) => (try Term.dest_Const) head) >>=
    (fn (c, T) => SOME c) 


fun all_consts_of_prop thm = Term.term_consts (prop_of thm)


fun set_opname1 (Const(nm, ty)) nnm = Const(nnm, ty)
  | set_opname1 (s $ t) nnm = (set_opname1 s nnm) $ t
  | set_opname1 _ nnm = error "Unexpected pattern in set_opname1" 
fun set_opname (Const("==>", ty) $ s $ t) nnm = (Const("==>", ty) $ s $ (set_opname t nnm))
  | set_opname (Const("==", ty) $ s $ t) nnm = (Const("==", ty) $ (set_opname1 s nnm)) $ t
  | set_opname (s $ t) nnm = (set_opname s nnm) $ t
  | set_opname _ nnm = error "Unexpected pattern in set_opname"



fun unq_path from to =
 let val fname = name_of_thy from
     fun path p thy = (if ((name_of_thy thy) = fname) then p
                       else (case (parents_of thy) of
                              [thy'] => (path (thy::p) thy')  
                            | _ => []))
 in
  path [] to
 end 




(* ************************************************************* *)
(*           PROOF RECONSTRUCTION AND THEOREM DERIVATION         *)
(* ************************************************************* *)

(* From the proof checker: look up a theorem.
 * Note that thm, which appears to do the same, but does not need a 
 * theory argument, depends on the elusive local context. 
 *)
fun lookup_thm thy =
  let val tab = foldr (uncurry(Symtab.update))
      (flat (map thms_of (thy :: Theory.ancestors_of thy)), 
       Symtab.empty)
  in (fn s => (uncurry(Symtab.lookup)) (tab, s))
  end

fun get_thm thy s = 
    (case lookup_thm thy s of 
	 NONE => (let val _ = Output.error_msg ("Unknown theorem " ^ quote s)
                  in
                   error "" 
                  end)
       | SOME thm => thm)

(* Freezes free variables in proof *)
fun freeze prf = 
      let val (pt2, _)= Proofterm.freeze_thaw_prf (prf) in pt2 end

(* Strip a PThm followed by applications of %;
 * ie. strip_thm PThm(..., prf, ...) % .. % ..  is prf.
 * Note that Proofterm.strip_thm also strips away outer %%, which seems
 * potentially harmful for our uses.
 *)
fun strip_thm prf = 
   (case Proofterm.strip_combt prf of
      (PThm (_, prf', _, _), _) => prf'
          |  _ => prf)

(* Accessor functions on theorem representations *)
fun get_prf thm  = #2 (#der (rep_thm thm))
(* ... and strip top Prf(..) *)
fun proof_of thm = strip_thm (get_prf thm)

fun proof_of' thm = (Thm.name_of_thm thm, proof_of thm)

(* Checks whether a theorem "name" occurs in a proof *)
fun occurs_thm name (PThm(("", _), p, _, _))= 
	if (name= "") then true else (occurs_thm name p)
  | occurs_thm name (PThm((thmnm, _), _, _, _)) = (name= thmnm)
  | occurs_thm name (Abst(_, _, p))= occurs_thm name p
  | occurs_thm name (AbsP(_, _, p))= occurs_thm name p
  | occurs_thm name (p % _)        = occurs_thm name p
  | occurs_thm name (p %% q)       = if (occurs_thm name p) 
					then true else (occurs_thm name q)
  | occurs_thm _    _              = false



fun all_occurs_thm_ nms (PThm(("", _), p, _, _))= all_occurs_thm_ nms p
  | all_occurs_thm_ nms (PThm((nm, _), _, _, _))= (let fun chk x [] = false
                                                         | chk x (y::ys) = ((x=y) orelse (chk x ys))
                                                   in
                                                    if (chk nm nms) then nms
                                                    else (nm::nms)
                                                   end)
  | all_occurs_thm_ nms (Abst(_, _, p)) = all_occurs_thm_ nms p
  | all_occurs_thm_ nms (AbsP(_, _, p)) = all_occurs_thm_ nms p
  | all_occurs_thm_ nms (p % _)         = all_occurs_thm_ nms p
  | all_occurs_thm_ nms (p %% q)        = (let val nms1 = all_occurs_thm_ nms p 
                                           in
                                            all_occurs_thm_ nms1 q
                                           end)
  | all_occurs_thm_ nms _ = nms

val all_occurs_thm = all_occurs_thm_ []



(* **************************************************** *
 * Necessary to expand reconstructed proofterm for      *
 * anonymous theorems (""), this is now done in one go  *
 * by functions "reconstruct_prf" and "fullprf_of".     *
 * **************************************************** *)

(* ISABELLE2003: modified argument to     *
 * expand_proof from [""] to [("", NONE)] *)
fun reconstruct_prf sg prop prf= 
   let val prf' = Reconstruct.reconstruct_proof sg prop prf
   in if (occurs_thm "" prf')
      then (Reconstruct.expand_proof sg [("", NONE)] prf')
      else prf' 
   end

(* Replaces build_prf 
fun fullprf_of thm = 
      reconstruct_prf (sign_of_thm thm) (prop_of thm) (proof_of thm);
*)


fun reconstruct_prf' sg prop nm prf = 
let val prf' = Reconstruct.reconstruct_proof sg prop prf
    val prf'' = (if (occurs_thm nm prf')
                  then (Reconstruct.expand_proof sg [(nm, NONE)] prf')
                  else prf')
in
  if (occurs_thm "" prf'') then (Reconstruct.expand_proof sg [("", NONE)] prf'')
                           else prf''  
end


fun fullprf_of thm = reconstruct_prf' (sign_of_thm thm) (prop_of thm) (Thm.name_of_thm thm) (proof_of thm)

fun fullprf_of' thm = 
      (Thm.name_of_thm thm, fullprf_of thm);


(* *********************************** *)
(* ***         derive_thm          *** *)
(* *********************************** *) 

val beta_eta_convert = Drule.fconv_rule Drule.beta_eta_conversion


(* Derive theorem from proofterm in given theory *)
fun derive_thm thy prf = 
     Drule.standard (varifyT (ProofChecker.thm_of_proof thy (freeze prf))) 
        handle e=> let val _ = writeln ("Exception in derive_thm: ")
                    in print_exn e end;



(* Verify that the proof of a theorem can be proofchecked *)
fun check_prf_of thy thm = derive_thm thy (fullprf_of thm)

(* ************************************************************* *)
(*     MAPPING AND FOLDING THEOREMS/AXIOMS IN PROOF TERMS        *)
(* ************************************************************* *)


(* These functions should be in proofterm.sml: *)

fun fold_terms_in_prf f (a, Abst(_, _, prf)) =  fold_terms_in_prf f (a, prf)
  | fold_terms_in_prf f (a, AbsP(_, SOME t, prf)) =
    fold_terms_in_prf f (f (t, a), prf)
  | fold_terms_in_prf f (a, AbsP(_, NONE, prf)) = fold_terms_in_prf f (a, prf)
  | fold_terms_in_prf f (a, PThm(_, prf, t, _)) = f (t, a)
  | fold_terms_in_prf f (a, PAxm(_, t, _))  = f (t, a)
  | fold_terms_in_prf f (a, prf1 %% prf2) = 
    fold_terms_in_prf f (fold_terms_in_prf f (a, prf1), prf2)
  | fold_terms_in_prf f (a, prf % SOME t) = f(t, fold_terms_in_prf f (a, prf))
  | fold_terms_in_prf f (a, Hyp t) = f(t, a)
  | fold_terms_in_prf f (a, MinProof ps) = (let val (l, _, _) = ps
                                            in
                                              foldl (fold_terms_in_prf f) (a, map snd l)
                                            end) 
  | fold_terms_in_prf f (a, _)     = a

fun fold_thms_in_prf f (a, Abst(_, _, prf)) = 
    fold_thms_in_prf f (a, prf)
  | fold_thms_in_prf f (a, AbsP(_, _, prf)) = 
    fold_thms_in_prf f (a, prf)
  | fold_thms_in_prf f (a, p as PThm _)  = f (a, p)
  | fold_thms_in_prf f (a, p as PAxm _)  = f (a, p)
  | fold_thms_in_prf f (a, prf1 %% prf2) = 
    fold_thms_in_prf f (fold_thms_in_prf f (a, prf1), prf2)
  | fold_thms_in_prf f (a, prf % _)      = fold_thms_in_prf f (a, prf)
  | fold_thms_in_prf f (a, _)            = a
    
fun map_thms_in_prf f (Abst(s, t, prf)) = Abst(s, t, map_thms_in_prf f prf)
  | map_thms_in_prf f (AbsP(s, t, prf)) = AbsP(s, t, map_thms_in_prf f prf)
  | map_thms_in_prf f (p as PThm _)     = f p
  | map_thms_in_prf f (p as PAxm _)     = f p
  | map_thms_in_prf f (prf1 %% prf2)    = map_thms_in_prf f prf1 %%
                                          map_thms_in_prf f prf2
  | map_thms_in_prf f (prf % t)         = map_thms_in_prf f prf % t
  | map_thms_in_prf f q                 = q


(* ************************************************************* *)
(*                  PROOF PRINTING FUNCTIONS                     *)
(* ************************************************************* *)

(* Print proof *)
fun print_prf sg prf = 
      writeln (Pretty.string_of 
       (ProofSyntax.pretty_proof sg (strip_thm prf)))

(* Print proof of a theorem *)
fun print_prf_thm thm = 
      print_prf (sign_of_thm thm) (get_prf thm)

(* ************************************************************* *)
(*                   FORWARD RESOLUTION                          *)
(* ************************************************************* *)

(* This is the forward resolution that we use to discharge proof obligations,
 * i.e. theorems lifted into the premises of the abstracted theorem: we 
 * first make all variables in the t into meta-quantified variables, 
 * because they will be in that form in the premises, then resolve with COMP 
 * (not RS).
 * We can't use RS since this lifts over the assumptions of the state as 
 * well, which in the case of induction is harmful.
 * On the other hand, we do want to lift over the params, but there seems
 * to be no tactic which does only that. The one below is a good guess, 
 * since we can rely on _all_ variables in the proof obligation to be bound.
 *)

fun comp  t s = (Drule.forall_intr_vars t) COMP s
fun comp' n s = (Drule.forall_intr_vars (thm n)) COMP s

(* Simplify premises (in HOL). See simple-example.sml how to use this. *)
fun simp t  = Seq.hd (ALLGOALS (full_simp_tac (simpset_of (theory "HOL"))) t)
fun simpi n t =  Seq.hd (full_simp_tac (simpset_of (theory "HOL")) n t)





end
