

signature INST_STRUCT =
sig
 type ExtInstStruct

 val add_type : Theory.theory -> string -> ExtInstStruct list -> (ExtInstStruct list * bool)

 val translate_structs : (string * mixfix option) Symtab.table -> Theory.theory -> 
                        (CType.CType Symtab.table -> Term.typ -> Term.typ) -> 
                         CType.CType Symtab.table -> Symtab.key Symtab.table ->
                         Derive_Thm.AxmImg Symtab.table -> ExtInstStruct list ->
                         (Theory.theory * CType.CType Symtab.table * Symtab.key Symtab.table *
                         Derive_Thm.AxmImg Symtab.table)

end  

structure Inst_Struct : INST_STRUCT =
struct

local open DatatypeAux in


datatype InstStruct = Datatype of {num : int,
                                   types : (int * (string list * string * mixfix)) list,
                                   constrs : (int * ((string * Term.typ * mixfix) list)) list}
                    | Record of {qname : string,
                                 recname : string,
                                 args : string list,
                                 arity : int,
                                 parent : (Term.typ list * string) option,
                                 fields : (string * Term.typ * mixfix) list}

type ExtInstStruct = {ids : string list, data : InstStruct, deps : string list}





fun try_datatype thy tid =
(case (DatatypePackage.datatype_info thy tid) of
  SOME {index, descr, induction, inject, distinct, ...} => 
   (let fun insert x xs = (if (x mem_string xs) then xs else (x::xs))
        fun un_str (a : string list, b) = a union b 
        fun diff_str (a : string list, b) = a \\ b 
        fun get_deps1 (DtTFree x) = []
          | get_deps1 (DtType (f, xs)) = insert f (Aux.foldl un_str ([], map get_deps1 xs))
          | get_deps1 (DtRec i) = []
        fun get_deps2 [] = []
          | get_deps2 ((c, args)::cs) = un_str(Aux.foldl un_str ([], map get_deps1 args), get_deps2 cs) 

        fun get_arity [] = []
          | get_arity (x::xs) = (case x of
                                  DtTFree v => v::(get_arity xs)
                                | _ => error ("Unexpected constructor in get_arity"))
        fun prep_constrs [] = []
          | prep_constrs ((Const (c, ty))::cs) = 
                                     (case (Syn_mng.get_mixfix c) of
                                           NONE => (c, Type.unvarifyT ty, NoSyn)::(prep_constrs cs)
                                         | SOME m => (c, Type.unvarifyT ty, m)::(prep_constrs cs))
          | prep_constrs (_ :: cs) = error ("Unexpected error in prep_constrs")

        fun get_data ids pdeps types constrs [] = (ids, pdeps, types, constrs)
          | get_data ids pdeps types constrs ((ind, (nm, args, cs))::dsc) = 
            let val mxfx = case (Syn_mng.get_type_mixfix nm) of
                             NONE => NoSyn
                           | SOME m => m
                val arity = get_arity args
                val deps = get_deps2 cs
                val pcs = (case (DatatypePackage.constrs_of thy nm) of
                              NONE => error ("Unexpected error: constrs_of " ^ nm)
                            | SOME cs_pr => prep_constrs cs_pr)
            in
              get_data (nm::ids) (un_str(deps, pdeps)) ((ind, (arity, nm, mxfx))::types) 
                       ((ind, pcs)::constrs) dsc
            end
        val (ids, pdeps, types, constrs) = get_data [] [] [] [] descr
        val deps = diff_str(pdeps, ids)
        val num = (length descr)  

        val data = Datatype{num = num, types = types, constrs = constrs}
        val ext = {ids = ids, data = data, deps = deps}
   in
      SOME ext
   end)
| NONE => NONE)



fun try_record thy tid =
      (case (try (unsuffix RecordPackage.ext_typeN) tid) of
         SOME x =>
                                    (* drop_base <thy>.<recname>.<recname>_ext_type = <thy>.<recname>
                                       drop_base <thy>.<recname>_scheme = <thy> *) 
          (case (RecordPackage.get_extension thy (NameSpace.drop_base tid)) of
              SOME info =>
                  (let val qname = NameSpace.drop_base tid
                       val recnm = NameSpace.base qname
                       val (a, i) = case (Aux.type_arity (thy, tid)) of
                                          NONE => error ("Unexpected exception with type_arity")
                                        | SOME (a, i) => (a, i) 


                       val (tc, args) = info
                       val src_rec_type = Type (tc, args)
                       val recT_fields = fst(RecordPackage.get_recT_fields thy src_rec_type)
                       
                       fun extract_args [] = error "Unexpected empty arguments for extract_args"
                         | extract_args [u] = [] 
                         | extract_args (x::xs) = (case x of
                                                    TFree (v, s) => (v::(extract_args xs))
                                                   | _ => error ("Unexpected constructor of a \
                                                                  \ record-type argument in extract_args"))
                       val arity = extract_args args
                       val parent = NONE 

                       fun insert x xs = (if (x mem_string xs) then xs else (x::xs))
                       fun un_str (a : string list, b) = a union b
                       fun un_str' (a : string list, b) = case b of NONE => a
                                                          | SOME x => a union x

                       fun extract_dep (Type (f, ts)) = insert f (Aux.foldl un_str ([], map extract_dep ts))
                         | extract_dep _ = []
                       fun extract_deps [] = []
                         | extract_deps ((sel, tp)::flds) = un_str(extract_dep tp, extract_deps flds)       

                       val deps = un_str' (extract_deps recT_fields, 
                                           Option.map (extract_dep o (fn (tp, f) => Type (f, tp))) parent) 

                       fun prep_fields [] = []
                         | prep_fields ((sel, tp)::flds) =
                           (case (Syn_mng.get_mixfix sel) of
                              NONE => (sel, tp, NoSyn)::(prep_fields flds)
                            | SOME mx => (sel, tp, mx)::(prep_fields flds))
                       val fields = prep_fields recT_fields

                       val recdata = Record {qname = qname, recname = recnm,
                                             args = arity, arity = i, parent = parent,
                                             fields = fields}
                       val extdata = {ids = [qname], data = recdata, deps = deps} 
                   in
                    SOME extdata
                   end)
            | NONE => error ("Unexpected error: no extension for " ^ x))
       | NONE => NONE)



fun add_type thy tid structs =
 let fun add r tid [] =
   (if !AWE.skip_datatypes
    then (if !AWE.skip_records 
          then (r, false)
          else (case (try_record thy tid) of
                      SOME data => (data::r, true)
                    | NONE => (r, false)))
    else (if !AWE.skip_records
          then (case (try_datatype thy tid) of
                 SOME data => (data::r, true)
               | NONE => (r, false))
          else  
          (case (try_datatype thy tid) of
            SOME data => (data::r, true)
          | NONE => (case (try_record thy tid) of
                      SOME data => (data::r, true)
                    | NONE => (r, false)))))      

       | add r tid ((x as {ids, ...} : ExtInstStruct)::xs) = 
         (if (tid mem_string ids) then (r @ (x::xs), true)
          else add (x::r) tid xs)
 in
  add [] tid structs
 end
 



fun translate_record renames trg map_type typeMap opMap axmMap 
         (Record{qname, recname, args, arity, parent, fields}) =
 let val trg_id = Aux.name_of_thy trg
     val recname' = (case (Symtab.lookup renames qname) of
                       NONE => NameSpace.base recname
                     | SOME (r, _) => r)
     val qname' = trg_id ^ "." ^ recname'

     fun map_parent (ts, p) = (let val ts' = map (map_type typeMap) ts
                                   val p' = (case (Symtab.lookup renames p) of
                                                 NONE => (trg_id ^ "." ^ (NameSpace.base p))
                                               | SOME (x, _) => (trg_id ^ "." ^ x))
                                  in
                                   (ts', p')
                                  end)

     val parent' = Option.map map_parent parent

     fun map_fields [] = []
       | map_fields ((sel, ty, mxfx')::xs) =
         let val (nsel, mxfx) = (case (Symtab.lookup renames sel) of
                                  NONE => (case (Syn_mng.get_mixfix sel) of
                                             NONE => (NameSpace.base sel, mxfx')
                                           | SOME m => (NameSpace.base sel, m))
                                | SOME (nm, mx) => (case mx of
                                                      NONE => (nm, mxfx')
                                                    | SOME m => (nm, m)))
             val _ = (case mxfx of
                       NoSyn => ()
                     | _ => Syn_mng.add_mixfix (qname' ^ "." ^ nsel, mxfx))
             val ty' = map_type typeMap ty
         in
           (nsel, ty', mxfx)::(map_fields xs)
         end
      val fields' = map_fields fields

      val tid = qname ^ "." ^ recname ^ RecordPackage.ext_typeN        
      val tid' = trg_id ^ "." ^ recname' ^ "." ^ recname' ^ RecordPackage.ext_typeN 

      val typeMap' = Symtab.update_new (tid, CType.CType_of_TC (tid', arity)) typeMap
                     handle DUP => error ("Unexpected DUP in translate_record, typeMap: " ^ tid)
      val opMap' = Symtab.update_new (qname ^ ".Abs_" ^ recname ^ RecordPackage.ext_typeN,
                                      qname' ^ ".Abs_" ^ recname' ^ RecordPackage.ext_typeN)
                   (Symtab.update_new (qname ^ ".Rep_" ^ recname ^ RecordPackage.ext_typeN,
                                       qname' ^ ".Rep_" ^ recname' ^ RecordPackage.ext_typeN) opMap)

                    handle DUP => error ("Unexpected DUP in translate_record, opMap")

      val axmMap' =  Symtab.update_new 
                      (qname ^ ".type_definition_" ^ recname ^ RecordPackage.ext_typeN,
                       Derive_Thm.Axm (qname' ^ ".type_definition_" ^ recname' ^ RecordPackage.ext_typeN)) axmMap
                    handle DUP => error ("Unexpected DUP in translate_record, axmMap")


      (* Adding record *)  
      val _ = Output.info ("... adding record " ^ qname ^ 
                          " from the domain as " ^ recname' ^ " to the codomain ...")
      val trg' = trg 
                 |> RecordPackage.add_record_i (args, recname') parent' fields'  

 in
  (trg', typeMap', opMap', axmMap')
 end



fun translate_datatype renames trg map_type typeMap opMap axmMap (Datatype{num, types, constrs}) =
 let val trg_id = Aux.name_of_thy trg

     fun extract_args typemap (Type ("fun", [l, t])) = (map_type typemap l)::(extract_args typemap t)
       | extract_args typemap _ = []

     fun prep_constrs tid typemap [] = []
       | prep_constrs tid typemap ((c, ty, mxfx)::cs) =
         (let val (c', mxfx') = (case (Symtab.lookup renames c) of
                                  SOME (r, m) => (case m of
                                                   NONE => (r, mxfx)
                                                 | SOME mx => (r, mx))
                                | NONE => (NameSpace.base c, mxfx))
              val args = (extract_args typemap ty)
              val _ = (case mxfx' of
                        NoSyn => ()
                      | _ => Syn_mng.add_mixfix (tid ^ "." ^ c', mxfx'))
          in
           (c', args, mxfx')::(prep_constrs tid typemap cs)
          end) 

     fun prep_types typemap res [] = (typemap, res)
       | prep_types typemap res ((ind, (arity, id, mxfx))::tys) =
         (let val (tid, mxfx') = 
                           (case (Symtab.lookup renames id) of
                             SOME (r, m) => (case m of
                                              NONE => (r, mxfx)
                                            | SOME mx => (r, mx))
                           | NONE => (NameSpace.base id, mxfx))
              val tid' = trg_id ^ "." ^ tid
              val _ = (case mxfx' of
                         NoSyn => ()
                       | _ => Syn_mng.add_type_mixfix (tid', mxfx'))

              val typemap' = Symtab.update_new (id, CType.CType_of_TC (tid', length arity)) typemap
          in
           prep_types typemap' ((ind, ((arity, tid, mxfx'), id))::res) tys
          end) 

     val (typeMap', types') = prep_types typeMap [] types

     fun prep opmap axmmap ids tids declrs [] = (opmap, axmmap, ids, tids, declrs)
       | prep opmap axmmap ids tids declrs (ind::is) = 
       (case (Aux.assoc_int (types', ind)) of
         NONE => error ("Unexpected index in translate_datatype #1")
       | SOME ((arity, tid, mxfx), id) => 
         (case (Aux.assoc_int (constrs, ind)) of
           NONE => error ("Unexpected index in translate_datatype #2")
         | SOME cs => (let val (q, base_id) = (NameSpace.drop_base id, NameSpace.base id)
                           val opmap' = Symtab.update_new (q ^ ".Abs_" ^ base_id,
                                                           trg_id ^ ".Abs_" ^ tid)
                                       (Symtab.update_new (q ^ ".Rep_" ^ base_id,
                                                           trg_id ^ ".Rep_" ^ tid) opmap)
                           
                           val axmmap' = Symtab.update_new (q ^ ".type_definition_" ^ base_id,
                                                   Derive_Thm.Axm (trg_id ^ ".type_definition_" ^ tid)) axmmap  
                           val tid' = trg_id ^ "." ^ tid
                           val cs' = prep_constrs tid' typeMap' cs
                           val decl = (arity, tid, mxfx, cs')
                       in
                        prep opmap' axmmap' (ids @ [id]) (tids @ [tid]) (declrs @ [decl]) is
                       end)))

     val inds = 0 upto (num - 1)      
     val (opMap', axmMap', ids, tids, declrs) = prep opMap axmMap [] [] [] inds 

 (* Adding datatype(s) *)  
     val _ = Output.info ("... adding datatype(s) " ^ (commas ids) ^ 
                          " from the domain as " ^ (commas tids) ^ " to the codomain ...")
     val (trg', _) = trg 
                     |> DatatypePackage.add_datatype_i false false tids declrs
 in
  (trg', typeMap', opMap', axmMap')
 end




fun translate_structs renames trg map_type typeMap opMap axmMap extdata =
  let val extdata' = Dependency.gen_trans_dep extdata
      val sort_data = Dependency.gen_sort extdata'

      fun translate trg typeMap opMap axmMap [] = (trg, typeMap, opMap, axmMap)
        | translate trg typeMap opMap axmMap (({data, ...})::xs) = 
          (case data of
            Datatype d => (let val (trg', typeMap', opMap', axmMap') =
                                    translate_datatype renames trg map_type typeMap opMap axmMap (Datatype d)
                           in
                            translate trg' typeMap' opMap' axmMap' xs
                           end)
          | Record r => (let val (trg', typeMap', opMap', axmMap') = 
                                 translate_record renames trg map_type typeMap opMap axmMap (Record r)
                         in
                          translate trg' typeMap' opMap' axmMap' xs
                         end))
  in
   translate trg typeMap opMap axmMap sort_data
  end 








end (* local open ... *)


end
